"""
Create data for SeaFlux filling data for surface pCO2 products
"""

import numpy as np
import pandas as pd
import xarray as xr

from fetch_data import download

from .utils import preprocess


def main(catalog_fname="../data/spco2_data.yml", dest="../data/output/"):
    """
    Runs functions to compute the filling data for SeaFlux
    """
    from .utils import save_seaflux

    ds = SOCOMensemble(catalog_fname).data

    # remove JMA for a longer time series (1985-2019)
    drop = ["seamask", "MPI_ULB_SOMFFN", "month", "JMA_MLR"]
    pco2 = ds.drop(drop).to_array("product")
    clim = ds.MPI_ULB_SOMFFN.sel(month=pco2.time.dt.month).drop("month")

    mask = pco2.notnull().all("product")
    pco2 = pco2.where(mask).dropna("time", "all")

    variable = "spco2filler"
    products = pco2.product.values.astype(str)
    reference_info = f"Ensemble of spco2 products: {', '.join(products)}"

    filler = make_filling_data(pco2, clim, "SOCOM ensemble", varname=variable)
    filler.attrs["climatology_data"] = "MPI-ULB-SOMFFN"
    filler.attrs["reference_data"] = reference_info

    sname = save_seaflux(filler, dest, variable)

    return sname


def calculate_fluxes(spco2, apco2, tempC, salt, preshPa, wind, ice):
    """
    TODO: might delete
    """
    import seaflux as sf

    fluxes = sf.flux_bulk(
        tempC,
        salt,
        spco2,
        apco2,
        preshPa,
        wind,
        kw_func=sf.gas_transfer_CO2.k_Wa92,
        kw_scaling=16,
    )

    xds = xr.Dataset()
    xds["fgco2"] = fluxes * (1 - ice)
    xds["apco2"] = apco2
    xds["ice"] = ice
    xds["area"] = sf.utils.area_grid()

    xds.fgco2.attrs["units"] = "gC/m2/day"
    xds.fgco2.attrs["long_name"] = "air-sea flux of CO2"
    xds.fgco2.attrs["description"] = (
        "Air-sea co2 flux where upward is positive: \n"
        "FCO2 = K0 * kw * (spco2 - apco2) * ice \n"
        "kw = Wanninkhof (1992) scaled to 16 cm/hr; \n"
        "K0 = Weiss (1974); \n"
        "apco2 = xCO2mbl * (P - pH2O), using pH2O from Dickson et al 2007; \n"
        "spco2 = data products in spco2; \n"
        "ice = (1 - sea ice fraction)\n"
        "ERA5 winds + MSLP, OSTIA SST, EN4 salinity were used as auxiliary data"
    )

    xds.apco2.attrs.pop("citation")
    xds.apco2.attrs["long_name"] = "atmospheric partial pressure of CO2"
    xds.apco2.attrs["units"] = "uatm"
    xds.apco2.attrs["source"] = "https://www.esrl.noaa.gov/gmd/ccgg/mbl/"
    xds.apco2.attrs["description"] = (
        "Atmospheric pCO2 was calculated with:\n"
        "pCO2 = xCO2mbl * (P - pH2O)\n\n"
        "xCO2mbl = NOAAs xCO2 marine boundary layer product (surface) interpolated linearly along latitudes and extrapolated along longitudes;\n"
        "P = atmospheric pressure from ERA5;\n "
        "pH2O = water vapour pressure calculated using Dickson et al. (2007);"
    )

    xds.ice.attrs["long_name"] = "sea ice fraction"
    xds.ice.attrs["units"] = "/100"
    xds.ice.attrs["description"] = "Sea ice fraction from the OSTIAv2 product"

    return xds


def make_filling_data(reference, climatology, reference_name, varname="spco2filler"):
    """
    Calculates the filling data based on a reference dataset and a climatology.

    The reference name will be used for metadata. Varname is to ensure that
    the correct variable name is assigned to the filling output.
    """
    from datetime import datetime as dt

    from ..area import get_area_from_dataset

    print("[SeaFlux] Calculating filler data from scaled climatology")
    ref = reference.dropna("time", "all")
    clim = climatology.dropna("time", "all")
    area = get_area_from_dataset(clim)

    dims = list(set(ref.dims) - set(["time"]))

    factor = (ref / clim).weighted(area).mean(dims)
    filler = clim * factor

    ds = xr.Dataset(
        attrs=dict(
            date=dt.today().strftime("%Y-%m-%d"),
            software="https://github.com/lukegre/SeaFlux",
            contact="Luke Gregor (luke.gregor@usys.ethz.ch)",
            description=(
                "A surface pCO2 product to fill marginal and high latitude "
                "seas for surface ocean pCO2 products that have pseudo-global "
                "coverage. This should be used for global flux integration "
                "calculations to ensure that the area being compared is truly "
                "global. The filling product is based on the MPI-ULB-SOMFFN "
                "product (Landschuetzer et al., 2020)."
            ),
        )
    )

    ds["scaling_factor"] = factor.assign_attrs(
        description=(
            "A scaling factor for surface ocean pCO2 to scale the MPI-ULB-SOMFFN "
            "climatological product. The scaling factor is calculated as the "
            f"ratio between a reference dataset ({reference_name}) and the "
            "climatology: (reference / climatology).mean([lat, lon]), where the "
            "average is weighted by pixel area. For more info see Fay et al. "
            "(2021)."
        )
    )

    ds[varname] = filler.assign_attrs(
        units="uatm",
        description=(
            "Scaled surface ocean pCO2 as estimated using the MPI-ULB-SOMFFN "
            "product. The climatological pco2 is scaled to a reference dataset "
            f"({reference_name}). The scaling is performed with: "
            "reference / climatology. For more details, see Fay et al. (2021)."
        ),
    )

    return ds


class SOCOMensemble:
    """
    Creates an object that makes it easy to access SOCOM data
    """

    def __init__(self, catalog_fname):
        """
        An object that downloads, reads in, homogenizes and combines
        surface ocean pCO2 products.

        The catalog must contain entries that match the functions in the
        SOCOM ensemble object. The catalog contains the urls to where the data
        can be downloaded. Further, the object contains functions are tailormade
        to homogenise the data sets so that they can easily be worked with.

        A default list of ensemble members is stored under ``self.members``
        which excludes the climatology and sea mask. The climatology name is
        stored under ``self.climatology``.

        The data can be accessed under self.data, which will load the dataset
        from memory if loaded, from a file if it exists, process each individual
        product, or download the products. A full pipeline!

        The catalog is not stored on GitHub as it contains passwords. Contact
        Luke for access to this catalog.
        """
        from fetch_data import read_catalog

        self.catalog_fname = catalog_fname
        self.cat = read_catalog(catalog_fname)

        self.members = [
            "mpi_somffn",
            "jena_mls",
            "cmems_ffnn",
            "csir_ml6",
            "not there",
            "nies_fnn",
            "jma_mlr",
        ]

        missing = [key for key in self.members if key not in self.cat]
        if any(missing):
            for m in missing:
                self.members.remove(m)

        self._data = None

        print("[SeaFlux] Default ensemble members in catalog:", ", ".join(self.members))

    @property
    def data(self):
        """quickly access the data as an object"""
        if self._data is None:
            self._data = self.__call__()
        return self._data

    def __call__(self, dest="../data/processed/spco2_socom_ensemble.nc"):
        """
        Will download all the data, process it and save it to the given destination
        """
        from pathlib import Path as path

        if path(dest).is_file():
            print(f"[SeaFlux] Returning existing file {dest}")
            return xr.open_dataset(dest)

        # mask
        mask = self.get_data("seamask").assign_attrs(long_name="ocean_mask")

        attrs = dict(units="uatm", long_name="surface_ocean_pco2")
        clim = self.get_data(self.climatology).assign_attrs(**attrs)

        ensemble = [self.get_data(k).assign_attrs(**attrs) for k in self.members]

        ds = xr.merge(ensemble).sel(time=slice("1982", None))
        ds = ds.assign_attrs(
            date_created=pd.Timestamp.today().strftime("%Y-%m-%d"),
            created_by="SeaFlux (Luke Gregor)",
            description=(
                "Surface ocean pCO2 from various data-based products. The information for these "
                "products can be found in each of the DataArrays's meta data. These data are "
                "part of the SOCOM ensemble and are updated annually. We limit the data from 1982 to "
                "present. "
            ),
        )
        ds[self.climatology.upper()] = clim
        ds["seamask"] = mask

        encode = {k: dict(zlib=True, complevel=4) for k in ds}
        ds.to_netcdf(dest, encoding=encode)

        return ds

    def get_data(self, name):
        """Helper function to download the data"""
        import logging

        if name not in self.cat:
            raise KeyError(
                f"{name} is not in the given data catalog {self.catalog_fname}"
            )

        func = getattr(self, f"get_{name}", None)
        if func is None:
            raise KeyError(f"get_{name} is not a function in {str(self)}")

        logging.info(f"[SeaFlux] retrieving {name.upper()}")

        kwargs = self.cat[name]
        meta = kwargs.get("meta", {})

        url = kwargs.get("url")
        url = url[0] if isinstance(url, list) else url

        data = func(kwargs).rename(name.upper()).load()

        data = data.assign_attrs(source=url, **meta)

        return data

    @staticmethod
    def get_mpi_ulb_somffn(entry):
        """processes data"""

        flist = download(**entry)

        xds = xr.open_mfdataset(flist)
        xda = xds.pco2.where(xds.pco2 > 0).coarsen(lat=4, lon=4).mean()
        xda = xda.rename("mpiulb_somffn").rename(time="month")

        pp = preprocess(rename_coordinates=False, center_months=False)

        xda = pp(xda)
        return xda

    @staticmethod
    def get_mpi_somffn(entry):
        """processes data"""

        flist = download(**entry)

        xda = xr.open_mfdataset(flist, drop_variables="date").spco2_raw
        xda = xda.rename("mpi_somffn").assign_attrs(
            units="uatm", source=entry["url"], **entry["meta"]
        )

        xda = preprocess()(xda)

        return xda

    @staticmethod
    def get_jena_mls(entry):
        """processes data"""

        flist = download(**entry)

        xds = xr.open_mfdataset(flist)
        xda = xds.pCO2.resample(mtime="1MS").mean("mtime")

        xda = xda.rename("jena_mls")
        xda = (
            xda.interp(
                lat=np.arange(-89.5, 90), lon=np.arange(-179.5, 180), method="nearest"
            )
            .roll(lon=180, roll_coords=False)
            .interpolate_na("lon", limit=20)
            .roll(lon=-180, roll_coords=False)
            .rename(mtime="time")
            .assign_attrs(
                units="uatm",
                source=entry["url"],
                **entry["meta"],
                history=(
                    "[Seaflux] resampled from daily to monthly and "
                    "interpolated to 1 degree using nearest neighbour interpolation"
                ),
            )
        )

        xda = preprocess()(xda)

        return xda

    @staticmethod
    def get_cmems_ffnn(entry):
        """processes data"""

        flist = download(**entry, verbose=20)

        xds = xr.open_mfdataset(flist, combine="nested", concat_dim="time")
        xda = (
            (xds.spco2 * 9.867)
            .assign_coords(longitude=(xds.longitude - 180) % 360 - 180)
            .rename(latitude="lat", longitude="lon")
            .resample(time="1MS")
            .mean()
            .sortby("lon")
            .assign_attrs(units="uatm", source=entry["url"], **entry["meta"])
        )

        xda = preprocess()(xda)

        return xda

    @staticmethod
    def get_nies_fnn(entry):
        """processes data"""
        from warnings import filterwarnings

        from ..fco2_pco2_conversion import fCO2_to_pCO2
        from ..utils import add_history
        from .aux_vars import download_era5_slp, download_sst_ice

        filterwarnings("ignore", category=RuntimeWarning)

        def decode_time(xds):
            """processes data"""
            import pandas as pd

            from datetime_matcher import DatetimeMatcher

            re_date = DatetimeMatcher()

            fname = xds.encoding["source"]
            datetime = re_date.extract_datetime("flux.%Y.ver", fname)
            year = pd.Timestamp(datetime).year

            y0, y1 = str(year), str(year + 1)
            time = pd.date_range(y0, y1, freq="1MS", closed="left")

            xds = xds.rename(month="time").assign_coords(time=time)

            xds = add_history(xds, "decode times manually")

            return xds

        flist = download(**entry)
        xda = xr.open_mfdataset(flist, preprocess=preprocess(decode_time)).fco2

        t0, t1 = [str(s) for s in xda.time.values[[0, -1]]]
        sst = xr.open_dataset(download_sst_ice())["sst"].sel(time=slice(t0, t1))
        msl = xr.open_dataset(download_era5_slp())["sp"].sel(time=slice(t0, t1)) / 100

        pco2 = xr.DataArray(
            fCO2_to_pCO2(xda, sst, msl),
            coords=xda.coords,
            dims=xda.dims,
            attrs=dict(units="uatm", source=entry["url"], **entry["meta"]),
        )

        pco2 = add_history(
            pco2, "re-shaped data from [year month lat lon] to [time lat lon]."
        )
        pco2 = add_history(
            pco2, "converted fCO2 to pCO2 using OISST v2.1, and ERA5 MSLP"
        )

        return pco2

    @staticmethod
    def get_jma_mlr(entry):
        """processes data"""

        def decode_time(xds):
            """processes data"""
            import pandas as pd

            from seaflux.data.utils import add_history

            time = xds.time
            unit = time.attrs.get("units")
            year = pd.to_datetime(unit.split()[-1]).year

            y0, y1 = str(year), str(year + 1)
            time = pd.date_range(y0, y1, freq="1MS", closed="left")
            xds = xds.assign_coords(time=time)

            xds = add_history(xds, "decode times manually")

            return xds

        flist = download(**entry, verbose=20)

        xda = xr.open_mfdataset(
            flist, decode_times=False, preprocess=preprocess(decode_time)
        ).pCO2s
        xda = xda.assign_attrs(units="uatm")

        return xda

    @staticmethod
    def get_csir_ml6(entry):
        """processes data"""

        flist = download(**entry, verbose=20)
        xds = xr.open_mfdataset(flist, preprocess=preprocess())

        xds = xds["spco2"].assign_attrs(
            units="uatm", source=entry["url"], **entry["meta"]
        )

        return xds

    @staticmethod
    def get_seamask(entry):
        """processes data"""

        flist = download(**entry)
        xds = xr.open_mfdataset(flist, preprocess=preprocess())

        xda = xds.seamask.assign_attrs(**entry["meta"])

        return xda



"""
Conversions of fCO2 - pCO2
--------------------------
"""


def fCO2_to_pCO2(fCO2SW_uatm, tempSW_C, pres_hPa=1013.25, tempEQ_C=None):
    """Convert fCO2 to pCO2 in sea water.

    If equilibrator temperature is provided, we get a simple approximate for
    equilibrator  :math:`xCO_2` that allows for the virial expansion to be
    calculated more accurately. If not, then a simple approximation is good
    enough. See the examples for the differences.

    .. math::
        pCO_2^{sw} = fCO_2^{sw} \\div virial(xCO_2^{eq})

    where :math:`xCO_2^{eq} = fCO_2^{sw} \\times \\Delta T^{(sw - eq)} \\div P^{eq}`


    Args:
        fCO2SW_uatm (array): seawater fugacity of CO2 in micro atmospheres
        tempSW_C (array): sea water temperature in degrees C
        pres_hPa (array, optional): equilibrator pressure in hecto Pascals. Defaults to 1013.25.
        tempEQ_C (array, optional): equilibrator temperature in degrees C. Defaults to None.

    Returns:
        array:
            partial pressure of CO2 in seawater

    Note:
        In FluxEngine, they account fully solve for the original xCO2 that is used
        in the calculation of the virial expansion. I use the first estimate of
        xCO2 (based on fCO2 rather than pCO2). The difference between the two
        approaches is so small that it is not significant to be concerned. Their
        correction is more precise, but the difference between their iterative
        correction and our approximation is on the order of 1e-14 atm (1e-8 uatm).

    Examples:
        >>> fCO2_to_pCO2(380, 8)
        381.50806485658234
        >>> fCO2_to_pCO2(380, 8, pres_hPa=985)
        381.4659553134281
        >>> fCO2_to_pCO2(380, 8, pres_hPa=985, tempEQ_C=14)
        381.466027968504
    """
    from . import auxiliary_equations as eqs
    from . import check_units as check

    # if equilibrator inputs are None, tempEQ=tempSW
    if tempEQ_C is None:
        tempEQ_was_None = True
        tempEQ_C = tempSW_C
    else:
        tempEQ_was_None = False

    # standardise the inputs and convert units
    fCO2sw = check.CO2_mol(fCO2SW_uatm * 1e-6)
    Tsw = check.temp_K(tempSW_C + 273.15)
    Teq = check.temp_K(tempEQ_C + 273.15)
    Peq = check.pres_atm(pres_hPa / 1013.25)

    # calculate the CO2 diff due to equilibrator and seawater temperatures
    # if statement is there to save a bit of time
    if tempEQ_was_None:
        dT = 1.0
    else:
        dT = eqs.temperature_correction(Tsw, Teq)

    # a best estimate of xCO2 - this is an approximation
    # one would have to use pCO2 / Peq to get real xCO2
    # Not getting the exact equilibrator xCO2
    xCO2eq = fCO2sw * dT / Peq

    pCO2SW = fCO2sw / virial_coeff(Tsw, Peq, xCO2eq)
    pCO2SW_uatm = pCO2SW * 1e6

    return pCO2SW_uatm


def pCO2_to_fCO2(pCO2SW_uatm, tempSW_C, pres_hPa=None, tempEQ_C=None):

    """Convert pCO2 to fCO2 in sea water to account for non-ideal behaviour of CO2

    If equilibrator temperature is provided, we get a simple approximate for
    equilibrator  :math:`xCO_2` that allows for the virial expansion to be
    calculated more accurately. If not, then a simple approximation is probably
    good enough. See the examples for the differences.

    .. math::
        pCO_2^{sw} = fCO_2^{sw} \\times virial(xCO_2^{eq})

    where :math:`xCO_2^{eq} = fCO_2^{sw} \\times \\Delta T^{(sw - eq)} \\div P^{eq}`

    Args:
        fCO2SW_uatm (array): seawater fugacity of CO2 in micro atmospheres
        tempSW_C (array): sea water temperature in degrees C
        pres_hPa (array, optional): equilibrator pressure in hecto Pascals. Defaults to 1013.25.
        tempEQ_C (array, optional): equilibrator temperature in degrees C. Defaults to None.

    Returns:
        array: fugacity of CO2 in seawater

    Note:
        In FluxEngine, they account for the change in xCO2. This error is so small
        that it is not significant to be concerned about it. Their correction is
        more precise, but the difference between their iterative correction and our
        approximation is less than 1e-14 atm (or 1e-8 uatm).

    Examples:
        >>> pCO2_to_fCO2(380, 8)
        378.49789637942064
        >>> pCO2_to_fCO2(380, 8, pres_hPa=985)
        378.53967828231225
        >>> pCO2_to_fCO2(380, 8, pres_hPa=985, tempEQ_C=14)
        378.53960618459695
    """
    from . import auxiliary_equations as eqs
    from . import check_units as check

    # if equilibrator inputs are None then make defaults Patm=1, tempEQ=tempSW
    if tempEQ_C is None:
        tempEQ_C = tempSW_C
    if pres_hPa is None:
        pres_hPa = 1013.25

    # standardise the inputs and convert units
    pCO2sw = check.CO2_mol(pCO2SW_uatm * 1e-6)
    Tsw = check.temp_K(tempSW_C + 273.15)
    Teq = check.temp_K(tempEQ_C + 273.15)
    Peq = check.pres_atm(pres_hPa / 1013.25)

    # calculate the CO2 diff due to equilibrator and seawater temperatures
    dT = eqs.temperature_correction(Tsw, Teq)
    # a best estimate of xCO2 - this is an approximation
    # one would have to use pCO2 / Peq to get real xCO2
    xCO2eq = pCO2sw * dT / Peq

    fCO2sw = pCO2sw * virial_coeff(Tsw, Peq, xCO2eq)
    fCO2sw_uatm = fCO2sw * 1e6

    return fCO2sw_uatm


def virial_coeff(temp_K, pres_atm, xCO2_mol=None):
    """
    Calculate the ideal gas correction factor for converting pCO2 to fCO2.

    Based on the Lewis and Wallace 1998 Correction.

    Args:
        temp_K (array): temperature in degrees Kelvin
        pres_atm (array): atmospheric pressure in atmospheres
        xCO2_mol (array, optional): mole fraction of CO2, can
            also be p/fCO2 if xCO2 not available. Can also be None
            which makes a small difference. See examples.

    Returns:
        array: the factor to multiply/divide with pCO2/fCO2. Unitless

        .. math::
            fCO_2 = pCO_2 \\times \\text{virial expansion}

            pCO_2 = fCO_2 \\div \\text{virial expansion}

    Examples:
        From Dickson et al. (2007)

        >>> 350 * virial_coeff(298.15, 1)  # CO2 [uatm] * correction factor
        348.8836492182758

    References:
        Weiss, R. (1974). Carbon dioxide in water and seawater: the solubility of a
        non-ideal gas. Marine Chemistry, 2(3), 203–215.
        https://doi.org/10.1016/0304-4203(74)90015-2

    Compared with the Seacarb package in R
    """
    from numpy import array, exp

    from . import check_units as check

    T = check.temp_K(temp_K)
    P = check.pres_atm(pres_atm)
    C = array(xCO2_mol)
    R = 82.057  # gas constant for ATM

    check.temp_K(T)
    check.pres_atm(P)

    # B is the virial coefficient for pure CO2
    B = -1636.75 + 12.0408 * T - 0.0327957 * T ** 2 + 3.16528e-5 * T ** 3
    # d is the virial coefficient for CO2 in air
    d = 57.7 - 0.118 * T

    # "x2" term often neglected (assumed = 1) in applications of Weiss's
    # (1974) equation 9
    if xCO2_mol is not None:
        check.CO2_mol(C)
        x2 = (1 - C) ** 2
    else:
        x2 = 1

    ve = exp(P * (B + 2 * x2 * d) / (R * T))

    return ve



"""
High level functions
--------------------


"""
from . import check_units as check
from . import gas_transfer_velocity
from . import solubility as sol


def flux_bulk(
    temp_bulk_C,
    salt_bulk,
    pCO2_bulk_uatm,
    pCO2_air_uatm,
    press_hPa,
    wind_ms,
    kw_func=gas_transfer_velocity.k_Ni00,
    as_dict=False,
):
    """
    Calculates bulk air-sea CO2 fluxes

    .. math::
        FCO_2 = k_w \\cdot K_0 \\cdot \\Delta pCO_2

    .. warning::
        Note that you have to be very aware of the impact on resampled winds
        where there is a loss of variability when averaging. For now, I recommend
        to use the lower level functions

    Args:
        temp_bulk_C (array): temperature from OISST in degCelcius with an allowable
            range of [-2:45]
        salt_bulk (array): salinity from EN4 in PSU. Allowable range [5 : 50]
        pCO2_bulk_uatm (array): partial pressure of CO2 in the sea in micro-atmospheres.
            Allowable range is [50 : 1000]
        pCO2_air_uatm (array): partial pressure of CO2 in the air in micro-atmospheres.
            Allowable range is [50 : 1000].
        press_hPa (array): atmospheric pressure in hecto-Pascals with allowable range [500:1500]
        wind_ms (array): wind speed in metres per second with an allowable range of [0 : 40]
        kw_func (callable): a function that returns the gas transfer velocity in cm/hr.
            The default is the gas transfer volicty as calculated by Ho et al. (2006).
            This is the preferred method of Goddijn-Murphy et al. (2016). Other functions
            are available in the `gas_transfer` class. If you'd like to use custom
            inputs must be wind speed (m/s) and temperature (degC) and output must
            be cm/hr.
        as_dict (bool, optional): If True, will return a dictionary where output is stored
            under ``data`` and meta is stored under ``attrs``. This dictionary can be passed
            to ``xr.DataArray`` for convenience.

    Returns:
        array:
            Sea-air CO2 flux where positive is out of the ocean and negative is
            into the ocean. Units are gC.m-2.day-1 (grams Carbon per metre squared
            per day)
    """
    from numpy import array

    press_atm = array(press_hPa) / 1013.25

    SSTfnd_C = array(temp_bulk_C)
    SSTfnd_K = SSTfnd_C + 273.15

    SSSfnd = array(salt_bulk)

    pCO2sea = array(pCO2_bulk_uatm) * 1e-6  # to atm
    pCO2air = array(pCO2_air_uatm) * 1e-6

    # checking units
    SSTfnd_K = check.temp_K(SSTfnd_K)
    SSSfnd = check.salt(SSSfnd)
    press_atm = check.pres_atm(press_atm)
    pCO2sea = check.CO2_mol(pCO2sea)
    pCO2air = check.CO2_mol(pCO2air)
    wind_ms = check.wind_ms(wind_ms)

    K0blk = sol.solubility_weiss1974(SSSfnd, SSTfnd_K, press_atm)

    # molar mas of carbon in g . mmol-1
    mC = 12.0108 * 1000  # (g . mol-1) / (mmol . mol-1)

    """unit analysis
    kw = (cm . hr-1) * hr . day-1 . cm-1 . m
    kw = m . day-1
    """
    kw = kw_func(wind_ms, SSTfnd_C) * (24 / 100)

    """ unit analysis
    flux = (m . day-1) .  (mol . L-1 . atm-1) . atm . (gC . mmol-1)
    flux = (m . day-1) . (mmol . m-3 . atm-1) . atm . (gC . mmol-1)
    flux = gC . m-2 . day-1
    """
    CO2flux_bulk = kw * K0blk * (pCO2sea - pCO2air) * mC

    if as_dict:
        kw_name = kw_func.__name__[2:]
        return dict(
            data=CO2flux_bulk,
            attrs=dict(
                units="gC . m^-2 . day^-1",
                description=f"sea-air CO2 fluxes calculated with {kw_name}",
                long_name="sea-air CO2 fluxes",
            ),
        )

    return CO2flux_bulk



from pathlib import Path as path


base = str(path(__file__).resolve().parent.parent)


def area(aux_catalog_fname, dest="../data/output/"):
    """Computes the area of the SeaFlux grid cells"""
    import xarray as xr

    from fetch_data import read_catalog

    from ..area import get_area_from_dataset
    from .utils import save_seaflux

    cat = read_catalog(aux_catalog_fname)

    ds = xr.open_mfdataset(download_sst_ice(cat["oisst_v2"])).sst.rename("temp")
    area = get_area_from_dataset(ds)

    sname = save_seaflux(area, dest, "area")
    return sname


def solubility(aux_catalog_fname, dest="../data/output/"):
    """Computes SeaFlux solubility from SST, Salt and Pres"""
    import xarray as xr

    from fetch_data import read_catalog

    from ..solubility import solubility_weiss1974
    from .utils import save_seaflux

    cat = read_catalog(aux_catalog_fname)

    ds = xr.merge(
        [
            xr.open_dataset(download_sst_ice(cat["oisst_v2"])).sst.rename("temp")
            + 273.15,
            xr.open_dataset(download_salinity(cat["en4_g10"])).salinity.rename("salt"),
            xr.open_dataset(
                download_era5_slp(download_dest=cat["era5_mslp"])
            ).sp.rename("mslp")
            / 101325,
        ]
    )

    time_mask = ds.to_array("tmp").notnull().all("tmp").any(["lat", "lon"])
    ds = ds.where(time_mask, drop=True)

    # unit analysis
    # mol / L / atm --> mol / m3 / uatm
    # mol . L-1 . atm-1 * (1e3L . m-3) * (1e-6 atm . uatm-1) = * 1e-3
    arr = solubility_weiss1974(ds.salt, ds.temp, press_atm=ds.mslp) * 1e-3

    sol = xr.DataArray(
        data=arr,
        coords=ds.temp.coords,
        dims=ds.temp.dims,
        attrs={
            "description": "CO2 solubility in seawater using the formulation of Weiss 1974",
            "units": "mol/m3/uatm",
            "long_name": "CO2 solubility in seawater",
        },
    )

    sname = save_seaflux(sol, dest, "sol")

    return sname


def save_ice_cover(catalog_fname, dest="../data/output/"):
    """Calculates SeaFlux sea ice cover as a fraction"""
    import xarray as xr

    from fetch_data import read_catalog

    from .utils import save_seaflux

    cat = read_catalog(catalog_fname)

    fname = download_sst_ice(cat["oisst_v2"])
    variable = "ice"
    ice = xr.open_mfdataset(fname)["icec"].rename(variable) / 100
    ice = ice.sel(time=slice("1982", None))

    sname = save_seaflux(ice, dest, variable)

    return sname


def download_era5_slp(
    year=list(range(1982, 2021)),
    download_dest="../data/raw/ERA5_mslp/",
    process_dest="../data/processed/era5_mslp_monthly.nc",
    **kwargs,
):
    """
    Shortcut for fetching era5 data. requires the `cdsapi` to be
    correctly set up (~/.cdsapi)
    Uses data from `reanalysis-era5-single-levels`
    Fetches data as monthly files to keep requests to a reasonable size
    """
    import logging
    import os

    from pathlib import Path

    import cdsapi
    import xarray as xr

    from joblib import Parallel, delayed
    from numpy import ndarray

    from .utils import preprocess

    if path(process_dest).is_file():
        return process_dest

    if isinstance(year, (list, tuple, ndarray)):
        logging.info(f"[DOWNLOAD]  downloading to: {download_dest}")
        inputs = [dict(year=y, download_dest=download_dest) for y in year]
        flist = Parallel(n_jobs=1)(
            delayed(download_era5_slp)(**input_dict) for input_dict in inputs
        )
        ds = xr.open_mfdataset(flist, preprocess=preprocess())
        ds.to_netcdf(
            process_dest, encoding={k: dict(zlib=True, complevel=4) for k in ds}
        )
        return ds

    year = str(year)
    sname = os.path.join(download_dest, f"ERA5_surfpress_monthly_{year}.nc")

    if os.path.isfile(sname):
        return sname

    Path(sname).parent.mkdir(exist_ok=True, parents=True)
    cds_client = cdsapi.Client()
    cds_client.retrieve(
        "reanalysis-era5-single-levels-monthly-means",
        {
            "product_type": "monthly_averaged_reanalysis",
            "format": "netcdf",
            "variable": "surface_pressure",
            "year": year,
            "month": [
                "01",
                "02",
                "03",
                "04",
                "05",
                "06",
                "07",
                "08",
                "09",
                "10",
                "11",
                "12",
            ],
            "time": ["00:00"],
        },
        sname,
    )

    return sname


def download_salinity(
    catalog_entry,
    process_dest="../data/processed/en4_salt_temp.nc",
):
    """Downloads salinity from MetOffice for 1982 until today"""
    import xarray as xr

    from fetch_data import download

    from .utils import preprocess

    if path(process_dest).is_file():
        return process_dest

    flist = download(**catalog_entry)
    ds = preprocess()(
        xr.open_mfdataset(paths=flist)[["salinity"]]
        .sel(depth=0, method="nearest")
        .drop("depth")
    )

    encode = {k: dict(zlib=True, complevel=4) for k in ds}
    ds.load().to_netcdf(process_dest, encoding=encode)

    return process_dest


def download_sst_ice(
    catalog_entry,
    process_dest="../data/processed/noaa_oisst_sst_icec.nc",
):
    """Downloads OISSTv2 data from NOAA"""
    import xarray as xr

    from fetch_data import download

    from .utils import preprocess

    if path(process_dest).is_file():
        return process_dest

    flist = download(**catalog_entry)

    ds = (
        xr.open_mfdataset(paths=flist, preprocess=preprocess())
        .where(lambda a: a.icec.notnull())
        .drop("time_bnds")
    )

    ds.to_netcdf(process_dest, encoding={k: dict(zlib=True, complevel=4) for k in ds})

    return process_dest



"""
Sanity checks to ensure that output is in the correct units
"""


class UnitError(Exception):
    """Raised when units are incorrect"""

    pass


def check_array_bounds(arr, lims, action="warn", name=""):
    """
    Checks that units are within the given limits. If not, then
    will raise/warn the user. Will always raise an error if more
    than half of the non-nan values are outside the limits.

    Parameters
    ----------
    arr : array-like
        The array that will be checked
    lims : tuple
        lower and upper limits of checks
        note that limits are exclusive (i.e. < and >, and not >=/<=)
    action: string
        raise - will raise an error and not continue
        warn - will throw a warning and mask values with nan
        quiet - same as warn, but without warning
        ignore - nothing will be done, but may result in bad data
    name: string
        if given, will inform the user of the name of the array
        to make debugging easier

    Return
    ------
    arr : array-like
        returns the array, but if warn or quiet, will be masked
        with nans
    """

    from warnings import warn

    from numpy import any, array, isnan, nan

    arr = array(arr, ndmin=1, dtype=float)
    if arr.size <= 2:
        return arr

    outside = (arr < lims[0]) | (arr > lims[1])

    non_nan_count = arr.size - isnan(arr).sum()
    half_outside = outside.sum() > (non_nan_count * 0.5)
    if half_outside:
        raise UnitError(
            f"More than half of the values in {name} are outside the limits "
            f"{str(lims)}. Check that input contains the correct units."
        )

    msg = (
        f"There are {outside.sum():d} values that do not fall within "
        f"the given limits {str(lims)}"
        f" of {name}"
        if name != ""
        else ""
    )

    if any(outside) & (action == "raise"):
        raise UnitError(msg)
    elif action == "warn":
        if any(outside):
            warn(msg, Warning)
        arr[outside] = nan
    elif action == "quiet":
        arr[outside] = nan
    elif action == "ignore":
        pass
    else:
        raise Exception("action must have raise/warn/quiet/ignore as inputs")

    return arr


def temp_K(temp_K):
    """Checks units for temp_K"""
    return check_array_bounds(
        arr=temp_K, lims=(270, 318.5), action="warn", name="temperature (K)"
    )


def pres_atm(pres_atm):
    """Checks units for pres_atm"""
    return check_array_bounds(
        arr=pres_atm, lims=(0.5, 1.5), action="warn", name="Pressure (atm)"
    )


def CO2_mol(CO2_mol):
    """Checks units for CO2_mol"""
    return check_array_bounds(
        arr=CO2_mol,
        lims=(5e-6, 0.08),
        action="warn",
        name="CO2 mole fraction (ppm)",
    )


def salt(salt):
    """Checks units for salt"""
    return check_array_bounds(
        arr=salt, lims=(0, 50), action="warn", name="Salinity (PSU)"
    )


def wind_ms(wind_ms):
    """Checks units for wind_ms"""
    return check_array_bounds(
        arr=wind_ms, lims=(0, 50), action="warn", name="Wind speed (m/s)"
    )



"""
The zenodo url version is stored here. Makes it easier to update when new files are uploaded.

TODO: future version may hold the file names here too.
"""


zenodo_url = "https://zenodo.org/record/4664653"

year_range = [1985, 2019]
version = "v2021.02"
contact = "luke.gregor@usys.ethz.ch"
citation = (
    "Fay, A. R., Gregor, L., Landschuetzer, P., McKinley, G. A., Gruber, N., "
    "Gehlen, M., Iida, Y., Laruelle, G. G., Roedenbeck, C., and Zeng, J.: "
    "Harmonization of global surface ocean pCO2 mapped products and their "
    "flux calculations; an improved estimate of the ocean carbon sink, Earth "
    "Syst. Sci. Data Discuss. [preprint], https://doi.org/10.5194/essd-2021-16, "
    "in review, 2021."
)



"""
CO2 solubility in seawater
--------------------------
"""

from . import check_units as check
from . import vapour_pressure as vapress


def solubility_weiss1974(salt, temp_K, press_atm=1):
    """Calculates the solubility of CO2 in sea water

    Used in the calculation of air-sea CO2 fluxes. We use the formulation by
    Weiss (1974) summarised in Wanninkhof (2014).

    Args:
        salt (array): salinity in PSU
        temp_K (array): temperature in deg Kelvin
        press_atm (array): pressure in atmospheres. Used in the solubility
            correction for water vapour pressure. If not given, assumed
            that press_atm is 1atm

    Returns:
        array: solubility of CO2 in seawater (:math:`K_0`) in mol/L/atm

    Examples:
        from Weiss (1974) Table 2 but with pH2O correction

        >>> solubility_weiss1974(35, 299.15)
        0.029285284543519093
    """

    from numpy import exp, log

    T = check.temp_K(temp_K)
    S = check.salt(salt)
    P = check.pres_atm(press_atm)

    # from table in Wanninkhof 2014
    a1 = -58.0931
    a2 = +90.5069
    a3 = +22.2940
    b1 = +0.027766
    b2 = -0.025888
    b3 = +0.0050578

    T100 = T / 100
    K0 = exp(
        a1 + a2 * (100 / T) + a3 * log(T100) + S * (b1 + b2 * T100 + b3 * T100 ** 2)
    )

    pH2O = vapress.weiss1980(S, T)
    K0 = K0 / (P - pH2O)

    # mol / L / atm --> mol / m3 / uatm
    # mol . L-1 . atm-1 * (L . m-3) * (atm . uatm-1)
    #                       1000    *   1e-6

    return K0  # units mol/L/atm



"""
Water vapour pressure
---------------------
"""

from . import check_units as check


def weiss1980(salt, temp_K):
    """Water vapour pressure of seawater after Weiss and Price (1980)

    For a given salinity and temperature using the methods
    defined in Weiss (1974)

    Args:
        salt (array): salinity in PSU
        temp_K (array): temperature in deg Kelvin

    Returns:
        array: sea water vapour pressure in atm (:math:`pH_2O`)

    Examples:
        >>> vapress_weiss1980(35, 25+273.15)  # tempC + 273.15
        0.03065529996317971

    References:
        Weiss, R. (1974). Carbon dioxide in water and seawater: the solubility
        of a non-ideal gas. Marine Chemistry, 2(3), 203–215.
        https://doi.org/10.1016/0304-4203(74)90015-2

        Weiss, R., & Price, B. a. (1980). Nitrous oxide solubility in water
        and seawater. Marine Chemistry, 8(4), 347–359.
        https://doi.org/10.1016/0304-4203(80)90024-9
    """
    from numpy import exp, log

    T = check.temp_K(temp_K)
    S = check.salt(salt)

    # Equation comes straight from Weiss and Price (1980)
    pH2O = exp(+24.4543 - 67.4509 * (100 / T) - 4.8489 * log(T / 100) - 0.000544 * S)

    return pH2O


def dickson2007(salt, temp_K):
    """Water vapour pressure of seawater after Dickson et al. (2007)

    Calculates :math:`pH_2O` at a given salinity and temperature using the
    methods defined in Dickson et al. (2007; CO2 manual)

    Parameters
    ----------
    salt : np.array
        salinity
    temp_K : np.array
        temperature in deg Kelvin

    Returns
    -------
    sea_vapress : np.array
        sea water vapour pressure in atm

    Examples
    --------
    >>> vapress_dickson2007(35, 298.15)  # from Dickson et al. (2007) Ch 5.3.2
    0.030698866245809465

    """
    from numpy import exp

    T = check.temp_K(temp_K)
    S = check.salt(salt)

    ###################################################
    # WATER VAPOUR PRESSURE FOR PURE WATER
    ###################################################
    # alpha coefficients from Wafner and Pruss, (2002)
    a1 = -7.85951783
    a2 = +1.84408259
    a3 = -11.7866497
    a4 = +22.6807411
    a5 = -15.9618719
    a6 = +1.80122502
    # critical points for water
    Pc = 22.064 / 101325.0e-6  # convert to atmosphers
    Tc = 647.096
    # zeta numbers correspond with alpha numbers
    z = 1 - T / Tc
    z1 = z
    z2 = z ** 1.5
    z3 = z ** 3
    z4 = z ** 3.5
    z5 = z ** 4
    z6 = z ** 7.5
    # vapour pressure of pure water
    pure_water = Pc * exp(
        (Tc / T) * (a1 * z1 + a2 * z2 + a3 * z3 + a4 * z4 + a5 * z5 + a6 * z6)
    )

    ###################################################
    # WATER VAPOUR PRESSURE FOR SEA WATER
    ###################################################
    # osmotic coefficients at 25C - Millero 1974
    c0 = +0.90799
    c1 = -0.08992
    c2 = +0.18458
    c3 = -0.07395
    c4 = -0.00221
    # total molality of dissolved species
    total_molality = 31.998 * S / (1e3 - 1.005 * S)
    B1 = total_molality * 0.5
    B2 = B1 ** 2
    B3 = B1 ** 3
    B4 = B1 ** 4
    osmotic_coeff = c0 + c1 * B1 + c2 * B2 + c3 * B3 + c4 * B4

    seawater = pure_water * exp(-0.018 * osmotic_coeff * total_molality)

    return seawater



def test_limits_warn():
    from seaflux.check_units import check_array_bounds
    import numpy as np

    arr = np.random.normal(0, 2, size=100)
    n_outside = (arr < -2).sum() + (arr > 2).sum()

    out = check_array_bounds(arr, lims=(-2, 2), action="quiet", name="test array")
    n_nans = np.isnan(out).sum()

    assert n_nans == n_outside


def test_limits_raise():
    from seaflux.check_units import check_array_bounds, UnitError
    import numpy as np

    arr = np.random.normal(0, 2, size=100)
    try:
        check_array_bounds(arr, lims=(-2, 2), action="raise", name="test array")
    except UnitError:
        pass



# Tests against PyCO2SYS (originally v1.6.0).
#
# Weiss (1974) CO2 solubility cannot be directly tested because PyCO2SYS evaluates this
# in /kg units while SeaFlux uses /l units.

import numpy as np
import PyCO2SYS as pyco2

import seaflux as sf


# Seed random number generator for reproducibility
rng = np.random.default_rng(7)
npts = 1000

# Set generic test conditions
salinity = np.array([*rng.uniform(low=0, high=50, size=npts - 1), 0])
temperature = np.array([-2, 0, *rng.uniform(low=-2, high=45, size=npts - 2)])  # deg-C
pCO2_or_fCO2 = rng.uniform(low=1, high=1000, size=npts)  # microatm
pres_atm = 1  # atm

# Convert units
temperature_K = temperature + 273.15  # K


def test_vapour_pressure_weiss1980():
    # Calculate vapour pressures of H2O
    vpf_seaflux = sf.vapour_pressure.weiss1980(salinity, temperature_K)
    vpf_pyco2 = 1 - pyco2.gas.vpfactor(temperature, salinity)
    # Compare results - should be virtually identical
    assert np.all(np.isclose(vpf_seaflux, vpf_pyco2, rtol=1e-12, atol=0))


def test_virial_coefficient():
    # Calculate fugacity factors
    fugfac_seaflux = sf.fco2_pco2_conversion.virial_coeff(temperature_K, pres_atm)
    fugfac_pyco2 = pyco2.sys(2100, 8.1, 2, 3, temperature=temperature)[
        "fugacity_factor"
    ]
    # Compare results
    assert np.all(np.isclose(fugfac_seaflux, fugfac_pyco2, rtol=0, atol=1e-7))


def test_pCO2_to_fCO2_conversion():
    # Convert pCO2 to fCO2
    fCO2_pyco2 = pyco2.sys(pCO2_or_fCO2, 8.1, 4, 3, temperature=temperature)["fCO2"]
    fCO2_seaflux = sf.pCO2_to_fCO2(pCO2_or_fCO2, temperature)
    # Compare results - not perfect agreement, but "good enough"(?) i.e. <0.01 microatm
    not_nan = ~np.isnan(fCO2_seaflux)
    assert np.all(
        np.isclose(fCO2_pyco2[not_nan], fCO2_seaflux[not_nan], rtol=0, atol=0.01)
    )


def test_fCO2_to_pCO2_conversion():
    # Convert fCO2 to pCO2
    pCO2_pyco2 = pyco2.sys(pCO2_or_fCO2, 8.1, 5, 3, temperature=temperature)["pCO2"]
    pCO2_seaflux = sf.fCO2_to_pCO2(pCO2_or_fCO2, temperature)
    # Compare results - not perfect agreement, but "good enough"(?) i.e. <0.01 microatm
    not_nan = ~np.isnan(pCO2_seaflux)
    assert np.all(
        np.isclose(pCO2_pyco2[not_nan], pCO2_seaflux[not_nan], rtol=0, atol=0.01)
    )



import seaflux as sf


def test_fCO2_to_pCO2_full():
    known_val = 381.466027968504
    assert sf.fCO2_to_pCO2(380, 8, pres_hPa=985, tempEQ_C=14) == known_val


def test_fCO2_to_pCO2_wo_press_tempEQ():
    assert sf.fCO2_to_pCO2(380, 8) == 381.50806485658234


def test_fCO2_to_pCO2_wo_tempEQ():
    assert sf.fCO2_to_pCO2(380, 8, pres_hPa=985) == 381.4659553134281


def test_pCO2_to_fCO2_wo_press_tempEQ():
    from seaflux import pCO2_to_fCO2

    known_value = 378.49789637942064
    assert pCO2_to_fCO2(380, 8) == known_value


def test_pCO2_to_fCO2_wo_tempEQ():
    from seaflux import pCO2_to_fCO2

    known_value = 378.53967828231225
    assert pCO2_to_fCO2(380, 8, pres_hPa=985) == known_value


def test_pCO2_to_fCO2_full():
    from seaflux import pCO2_to_fCO2

    known_value = 378.53960618459695
    assert pCO2_to_fCO2(380, 8, pres_hPa=985, tempEQ_C=14) == known_value



"""
Module to directly download the latest version of SeaFlux data from Zenodo
"""

from pathlib import Path as path

from . import utils


seaflux_data_path = path(utils.__file__).resolve().parent.absolute()
catalog_name = str(seaflux_data_path / "zenodo.yml")


def get_zenodo_catalog():
    """fetches the default catalog and returns as a dictionary. The dictionary
    is presented as a YAML file if you are using IPython/Jupyter"""
    import fetch_data as fd

    return fd.read_catalog(catalog_name)


_default_catalog = get_zenodo_catalog()
_key = list(_default_catalog.keys())[0]
_default_catalog = _default_catalog[_key]
_dest = _default_catalog.get("dest", None)


def get_seaflux_data(catalog_name=catalog_name, dest=_dest, n_jobs=1, verbose=False):
    """Downloads SeaFlux data from Zenodo using the default yaml file containing
    the paths to the latest SeaFlux data. The data is downloaded and then
    combined. You can create your own yaml file to customise the files you want
    to access."""

    from datetime import datetime as dt

    import fetch_data as fd
    import xarray as xr

    from . import config
    from .utils import preprocess

    cat = fd.read_catalog(catalog_name)
    key = list(cat.keys())[0]
    entry = cat[key]
    entry["dest"] = dest

    flist = fd.download(**entry, n_jobs=n_jobs, verbose=verbose)

    xds = xr.open_mfdataset(flist, preprocess=preprocess())
    xds = xds.assign_attrs(
        product_name="SeaFlux",
        product_version=config.version,
        date_accessed=dt.now().strftime("%Y-%m-%d"),
        contact=config.contact,
        **entry["meta"]
    )

    return xds



"""
Area calculations
-----------------

Calculates the area of pixels for a give grid input.
"""


def earth_radius(lat):
    """Calculate the radius of the earth for a given latitude

    Args:
        lat (array, float): latitude value (-90 : 90)

    Returns:
        array: radius in metres
    """
    from numpy import cos, deg2rad, sin

    lat = deg2rad(lat)
    a = 6378137
    b = 6356752
    r = (
        ((a ** 2 * cos(lat)) ** 2 + (b ** 2 * sin(lat)) ** 2)
        / ((a * cos(lat)) ** 2 + (b * sin(lat)) ** 2)
    ) ** 0.5

    return r


def area_grid(lat, lon, return_dataarray=False):
    """Calculate the area of each grid cell for given lats and lons

    Args:
        lat (array): latitudes in decimal degrees of length N
        lon (array): longitudes in decimal degrees of length M
        return_dataarray (bool, optional): if True returns xr.DataArray, else array

    Returns:
        array, xr.DataArray: area of each grid cell in meters

    References:
        https://github.com/chadagreene/CDT/blob/master/cdt/cdtarea.m
    """
    from numpy import cos, deg2rad, gradient, meshgrid

    ylat, xlon = meshgrid(lat, lon)
    R = earth_radius(ylat)

    dlat = deg2rad(gradient(ylat, axis=1))
    dlon = deg2rad(gradient(xlon, axis=0))

    dy = dlat * R
    dx = dlon * R * cos(deg2rad(ylat))

    area = dy * dx

    if not return_dataarray:
        return area
    else:
        from xarray import DataArray

        xda = DataArray(
            area.T,
            dims=["lat", "lon"],
            coords={"lat": lat, "lon": lon},
            attrs=dict(
                long_name="Area per pixel",
                units="m^2",
                description=(
                    "Area per pixel as calculated by SeaFlux. The non-"
                    "spherical shape of Earth is taken into account."
                ),
            ),
        )

        return xda


def get_area_from_dataset(dataarray, lat_name="lat", lon_name="lon"):
    """
    Calculate the grid cell area from a xr.Dataset or xr.DataArray.
    """
    da = dataarray
    x = da.lon.values
    y = da.lat.values

    area = area_grid(y, x, return_dataarray=True)

    return area



"""
Utilities for data set homogenisation, metadata, and plotting
"""


def save_seaflux(xds, dest, variable_name):
    """Helper function to save netCDF files with the correct meta data"""
    from datetime import datetime as dt
    from pathlib import Path as path

    import xarray as xr

    from .config import citation, contact, version, year_range

    if isinstance(xds, xr.DataArray):
        xds = xds.to_dataset(name=variable_name)
    elif not isinstance(xds, xr.Dataset):
        raise TypeError("xds must be a dataset or dataarray")

    assert variable_name in xds
    if "time" in xds:
        xds = xds.sel(time=slice(str(year_range[0]), str(year_range[1])))
        t0, t1 = xds.time.dt.year.values[[0, -1]]
        date_range = f"_{t0}-{t1}"
    else:
        date_range = ""

    sname = f"SeaFlux_{version}_{variable_name}{date_range}.nc"
    full_path = path(dest) / sname

    xds.attrs["citation"] = citation
    xds.attrs["prodct_version"] = version
    xds.attrs["date_created"] = dt.now().strftime("Y%-%m-%d")
    xds.attrs["contact"] = contact

    print(f"[SeaFlux] Saving {variable_name} to {full_path}")

    encoding = {k: dict(zlib=True, complevel=4) for k in xds}
    xds.to_netcdf(full_path, encoding=encoding)

    return full_path


def add_history(xds, message):
    """
    Adds history to xr.Datasets with a time stamp and [SeaFlux] prefix
    """
    import pandas as pd

    time = pd.Timestamp.today().strftime("%Y-%m-%d")
    message = message.replace("'", "").replace('"', "")

    history = xds.attrs.get("history", "").strip()
    history = history if history.endswith(";") or history == "" else history + "; "
    history += f" [SeaFlux @ {time}] {message}; "

    xds = xds.assign_attrs(history=history.strip())

    return xds


def get_kwargs():
    """
    gets keyword=value pairs for a function and returns as a dict
    """
    import inspect

    frame = inspect.currentframe().f_back
    keys, _, _, values = inspect.getargvalues(frame)
    kwargs = {}
    for key in keys:
        if key != "self":
            kwargs[key] = values[key]
    return kwargs


def rename_coords(
    xds,
    lat=["latitude", "Lat", "ylat", "y", "Y"],
    lon=["longitude", "Lon", "xlon", "x", "X"],
    time=["Time", "tmnth", "t", "T"],
    region=["reg", "regions"],
    **kwargs,
):
    """renames coordinates to time, lat, lon from given presents"""

    coord_names = dict(
        lat=lat,
        lon=lon,
        time=time,
        region=region,
        **kwargs,
    )

    rename = {}
    for key in xds.coords:
        for new, old in coord_names.items():
            if key in old:
                rename[key] = new

    xds = xds.rename(**rename)

    if rename != {}:
        xds = add_history(xds, f"renamed variables to match protocol {str(rename)}")

    return xds


def order_lon_180(xds):
    """flips longitudes to -180 : 180"""
    if "lon" not in xds.dims:
        return xds

    lon = xds.lon.copy().values
    xds = xds.assign_coords(lon=(lon - 180) % 360 - 180).sortby("lon")

    if any(lon != xds.lon.values):
        xds = add_history(xds, "Flipped longitudes -180:180")

    return xds


def sort_lats(xds):
    """sort lats"""
    if "lat" not in xds.dims:
        return xds

    import numpy as np

    lat = xds.lat.values
    delta = np.diff(lat[[0, -1]])
    if delta < 0:
        xds = xds.sortby("lat")
        xds = add_history(xds, "Sorted lats")

    return xds


def interpolate_coords(xds):
    """interpolates coordinates onto a global grid"""
    import numpy as np

    dims = []
    if "lat" in xds.dims:
        dims += ("lat",)
    if "lon" in xds.dims:
        dims += ("lon",)

    must_interpolate = []
    for dim in dims:
        coords = xds[dim].values
        if ((coords + 0.5) % 1).sum():
            must_interpolate += (dim,)
    if must_interpolate == []:
        return xds

    if "lon" in must_interpolate:
        x = np.arange(-179.5, 180)
        xds = xds.interp(lon=x)
        xds = add_history(xds, "Interpolated `lon` onto grid centers")
    if "lat" in must_interpolate:
        y = np.arange(-89.5, 90)
        xds = xds.interp(lat=y)
        xds = add_history(xds, "Interpolated `lat` onto grid centers")

    return xds


def transpose(xds):
    """transpose data to [other], time, lat, lon"""
    reccap_order = ["time", "lat", "lon"]
    coords = list(xds.dims)

    if reccap_order == coords:
        return xds

    order = []
    for key in reccap_order:
        if key in coords:
            order += (key,)
            coords.remove(key)
    order = coords + order
    xds = xds.transpose(*order)
    xds = add_history(xds, f"Transposed dimensions to ({str(order)[1:-1]})")

    return xds


def center_time_on_15th(xds):
    """center monthly data on the 15th of the month"""
    import numpy as np
    import xarray as xr

    if "time" not in xds.dims:
        return xds

    dt = np.timedelta64(14, "D")
    t = xds.time.values.astype("datetime64[M]") + dt

    xds["time"] = xr.DataArray(data=t, dims=["time"], coords={"time": t})

    xds = add_history(xds, "Time set to 15th of each month")

    return xds


def add_coord_attrs(xds):
    """Ensures that files are Georeferened for Panoply"""
    xds["lat"].attrs = dict(
        standard_name="latitude",
        units="degrees_north",
        axis="Y",
    )
    xds["lon"].attrs = dict(
        standard_name="longitude",
        units="degrees_east",
        axis="X",
    )
    return xds


def preprocess(
    *args,
    rename_coordinates=True,
    sort_latitudes=True,
    center_months=True,
    interpolate_coordinates=True,
    lon_0_180=True,
    transpose_dims=True,
    add_coord_attributes=True,
):
    """
    Args are additional functions that are applied in sequence
    A function generator that can be used as follows:
        reccap2_formatted_xds = xr.open_mfdataset(
            fname, decode_times=False,
            preprocess=data.preprocess(decode_times=True))

    There are several options that can be switched on or off with boolean.
    All changes are documented and appended to xds.attrs.history
    """

    def dataprep(ds):
        "wrapped function"

        for func in args:
            ds = func(ds)

        if rename_coordinates:
            ds = rename_coords(ds)
        if transpose_dims:
            ds = transpose(ds)
        if sort_latitudes:
            ds = sort_lats(ds)
        if lon_0_180:
            ds = order_lon_180(ds)
        if interpolate_coordinates:
            ds = interpolate_coords(ds)
        if center_months:
            ds = center_time_on_15th(ds)
        if add_coord_attributes:
            ds = add_coord_attrs(ds)

        return ds

    return dataprep


def subplot_map(pos=111, proj=None, round=True, land_color="w", **kwargs):
    """
    Makes an axes object with a cartopy projection for the current figure

    Parameters
    ----------
    pos: int/list [111]
        Either a 3-digit integer or three separate integers
        describing the position of the subplot. If the three
        integers are *nrows*, *ncols*, and *index* in order, the
        subplot will take the *index* position on a grid with *nrows*
        rows and *ncols* columns. *index* starts at 1 in the upper left
        corner and increases to the right.
        *pos* is a three digit integer, where the first digit is the
        number of rows, the second the number of columns, and the third
        the index of the subplot. i.e. fig.add_subplot(235) is the same as
        fig.add_subplot(2, 3, 5). Note that all integers must be less than
        10 for this form to work.
    proj: crs.Projection()
        the cartopy coord reference system object to create the projection.
        Defaults to crs.PlateCarree(central_longitude=0) if not given
    round: bool [True]
        If the projection is stereographic, round will cut the corners and
        make the plot round
    land_color: str ['w']
        the color of the land patches
    **kwargs:
        passed to fig.add_subplot(**kwargs)

    Returns
    -------
    An axes object attached to the current figure. If no figure, a figure
    will be created as is default for plt.gcf()
    """
    import matplotlib.path as mpath
    import matplotlib.pyplot as plt
    import numpy as np

    from cartopy import crs, feature

    proj = crs.PlateCarree(0) if proj is None else proj

    fig = plt.gcf()
    ax = fig.add_subplot(pos, projection=proj, **kwargs)

    # makes maps round
    stereo_maps = (
        crs.Stereographic,
        crs.NorthPolarStereo,
        crs.SouthPolarStereo,
    )
    if isinstance(ax.projection, stereo_maps) & round:

        theta = np.linspace(0, 2 * np.pi, 100)
        center, radius = [0.5, 0.5], 0.475
        verts = np.vstack([np.sin(theta), np.cos(theta)]).T
        circle = mpath.Path(verts * radius + center)

        ax.set_boundary(circle, transform=ax.transAxes)

    # adds features
    ax.add_feature(feature.LAND, color=land_color, zorder=4)
    ax.add_feature(feature.COASTLINE, lw=0.5, zorder=4)
    ax.outline_patch.set_lw(0.5)
    ax.outline_patch.set_zorder(5)

    return ax


def style_line_subplot(ax, add_zero_line=True, xlim=None, y_range=None):
    """Styles line plots to look ready for publication"""
    import matplotlib as mpl
    import matplotlib.pyplot as plt
    import numpy as np

    mpl.rcParams["axes.spines.right"] = False
    mpl.rcParams["axes.spines.top"] = False

    if ax is None:
        ax = plt.gca()
    plt.sca(ax)

    plt.xticks(rotation=0, ha="center")
    plt.xlabel("")
    plt.title("")

    if ax.get_ylim()[0] < 0 < ax.get_ylim()[1]:
        ax.axhline(0, color="k", ls="--", lw=0.5, zorder=0)

    if xlim is None:
        xlim = ax.get_xlim()
    ax.set_xticks(np.arange("1980", "2020", 5, dtype="datetime64[Y]"))
    ax.set_xticklabels(np.arange(1980, 2020, 5))
    ax.set_xlim(*xlim)

    if y_range is not None:
        center = np.mean(ax.get_ylim())
        upp = center + y_range / 2
        low = center - y_range / 2
        ax.set_ylim(low, upp)

    return ax



"""
Additional equations
--------------------

Not necessarily linked to the marine carbonate system, but are 
useful. 
"""

from . import check_units as check
import numpy as np


def pressure_height_correction(pres_hPa, tempSW_C, sensor_height=10.0):
    """Returns exact sea level pressure if the sensor is measuring at height

    Args:
        pres_hPa (array): Pressure in kiloPascal measured at height
        tempSW_C (array): Temperature of the seawater in deg C
        sensor_height (float, optional): the height of the sensor above sea 
            level. Can be negative if you want to convert SLP to sensor height
            pressure. Defaults to 10.0.

    Returns:
        array: height corrected pressure
    """    

    T = check.temp_K(tempSW_C + 273.15)  # temperature in Kelvin
    P = check.pres_atm(pres_hPa / 1013.25)  # pressure in Pascal

    # Correction for pressure based on sensor height
    R = 8.314  # universal gas constant (J/mol/K)
    M = 0.02897  # molar mass of air in (kg/mol) - Wikipedia
    # Density of air at a given temperature. Here we assume
    # that the air temp is the same as the intake temperature
    d = P / (R / M * T)
    g = 9.8  # gravity in (m/s2)
    h = -sensor_height  # height in (m)
    # correction for atmospheric
    press_height_corr_hpa = (P - (d * g * h)) / 100.0

    return press_height_corr_hpa


def temperature_correction(temp_in, temp_out):
    """pCO2 correction factor for temperature changes

    Calculate a correction factor for the temperature difference between the
    intake and equilibrator. This is based on the empirical relationship used
    in Takahashi et al. 1993.

    .. math::

        pCO_2^{Tout} = pCO_2^{Tin} * T^{factor}

    Args:
        temp_in (array): temperature at which original pCO2 is measured
        temp_out (array): temperature for which pCO2 should be represented

    Returns:
        array: a correction factor to be multiplied to pCO2 (unitless)

    References:
        Takahashi, Taro et al. (1993). Seasonal variation of CO2 and nutrients in
        the high-latitude surface oceans: A comparative study. Global
        Biogeochemical Cycles, 7(4), 843–878. https://doi.org/10.1029/93GB02263
    """
    # see the Takahashi 1993 paper for full description

    Ti = np.array(temp_in)
    To = np.array(temp_out)

    factor = np.exp(0.0433 * (To - Ti) - 4.35e-05 * (To ** 2 - Ti ** 2))

    return factor



"""
Set up module access for the base package
"""
from . import auxiliary_equations as eqs
from . import gas_transfer_velocity as kw
from .area import get_area_from_dataset
from .data.download_zenodo_files import get_seaflux_data, get_zenodo_catalog
from .fco2_pco2_conversion import fCO2_to_pCO2, pCO2_to_fCO2
from .flux_calculations import flux_bulk



"""
Atmospheric pCO2
----------------
This module is not imported by default and has to be manually imported.
This is because it is used only in production of the SeaFlux atmospheric dataset

Contains functions for the full process:
1. Download the NOAA marine boundary layer product (done)
2. Interpolate the product onto a standard grid (done)
3. Download MSLP (ERA5), SST (AVHRR), salinity (EN4) for ATM calculations (done)
4. Convert atmospheric xCO2 to pCO2 with the data
5. Interpolate the final year (2020)

Note that there are no tests for this code, so it is very likely to break
"""

from pathlib import Path as path


base = str(path(__file__).resolve().parent.parent)


def main(
    noaa_mbl_url,
    download_dest="../data/raw/",
    aux_catalog_name="../data/aux_data.yml",
    processed_dest="../data/processed/",
    output_dest="../data/output/",
):
    """to be called when creating the atmospheric pCO2"""
    import xarray as xr

    from fetch_data import read_catalog
    from pandas import Timestamp

    from .aux_vars import download_era5_slp, download_salinity, download_sst_ice
    from .utils import center_time_on_15th, preprocess, save_seaflux

    if path(output_dest).is_file():
        return output_dest

    cat = read_catalog(aux_catalog_name)

    salt = download_salinity(cat["en4_g10"], f"{processed_dest}/en4_salt_temp.nc")
    temp = download_sst_ice(cat["oisst_v2"], f"{processed_dest}/noaa_oisst_sst_icec.nc")
    pres = download_era5_slp(
        download_dest=cat["era5_mslp"]["dest"],
        process_dest=f"{processed_dest}/era5_mslp_monthly.nc",
    )

    ds = xr.merge(
        [
            xr.open_dataset(salt)["salinity"].rename("saltPSU"),
            xr.open_dataset(temp)["sst"].rename("tempC"),
            xr.open_dataset(pres)["sp"].rename("presPa"),
        ]
    )

    noaa_mbl_xco2 = (
        download_noaa_mbl(
            noaa_mbl_url,
            download_dest=f"{download_dest}/co2_GHGreference_surface.txt",
            target_lat=ds.lat.values,
            target_lon=ds.lon.values,
        )
        .resample(time="1MS")
        .mean()
    )

    t0, t1 = ds.time.values[[0, -1]]
    noaa_mbl_xco2 = center_time_on_15th(noaa_mbl_xco2).sel(time=slice(t0, t1))

    t0, t1 = noaa_mbl_xco2.time.values[[0, -1]]
    ds = ds.sel(time=slice(t0, t1))

    atm_pco2 = atm_xCO2_to_pCO2(
        noaa_mbl_xco2, ds.presPa.where(ds.tempC.notnull()) / 100, ds.tempC, ds.saltPSU
    )

    atm_pco2 = preprocess()(
        xr.DataArray(
            data=atm_pco2,
            dims=ds.tempC.dims,
            coords=ds.tempC.coords,
            name="pco2atm",
            attrs=dict(
                long_name=(
                    "partial_pressure_of_carbon_dioxide_in_the_marine_boundary_layer"
                ),
                short_name="pco2atm",
                units="uatm",
                description=(
                    "Atmospheric pCO2 for the marine boundary layer is calculated "
                    "from the NOAAs marine boundary layer pCO2 with: xCO2 * (Patm "
                    "- pH2O). Where pH2O is calculated using vapour pressure from "
                    "Dickson et al. (2007)"
                ),
                history=(
                    getattr(noaa_mbl_xco2, "history", "").strip(";") + ";\n"
                    f"[SeaFlux @ {Timestamp.today():%Y-%m-%d}] "
                    f"pCO2 calculated from xCO2 * (Patm - pH2O), where "
                    f"pH2O is calculated with Dickson et al. (2007)"
                ),
                citation=(
                    "Ed Dlugokencky and Pieter Tans, NOAA/ESRL "
                    "(www.esrl.noaa.gov/gmd/ccgg/trends/)"
                ),
            ),
        )
    )

    variable = "pco2atm"
    pco2atm = interpolate_year(atm_pco2).to_dataset(name=variable)

    sname = save_seaflux(pco2atm, output_dest, variable)

    return sname


def atm_xCO2_to_pCO2(xCO2_ppm, slp_hPa, tempSW_C, salt):
    """
    Convert atmospheric xCO2 to pCO2 with correction for water vapour pressure
        pCO2atm = xCO2atm * (Press - pH2O)

    Args:
        xCO2_ppm (array): atmospheric, or marine boundary layer mole fraction of CO2 (NOAA MBL)
        slp_hPa (array): sea water temperature in degrees C (ERA5 recommended)
        tempSW_C (array): atmospheric pressure in hecto Pascal (NOAA AVHRR OISSTv2 recommended)
        salt (array): sea surface salinity in PSU (EN4 salinity)

    Returns:
        array: note that output will be an np.ndarray regardless of input
    """
    from numpy import array

    from .. import check_units as check
    from .. import vapour_pressure as vapress

    print("[SeaFlux] Converting xCO2 to pCO2")
    xCO2 = array(xCO2_ppm)
    # check units and mask where outsider of range
    Tsw = check.temp_K(tempSW_C + 273.15)
    Ssw = check.salt(salt)
    Patm = check.pres_atm(slp_hPa / 1013.25)

    pH2O = vapress.dickson2007(Ssw, Tsw)

    pCO2atm = xCO2 * (Patm - pH2O)

    return pCO2atm


def read_noaa_mbl_url(noaa_mbl_url, dest):
    """Downloads url and reads in the MBL surface file

    Args:
        noaa_mbl_url (str): the address for the noaa surface file
        dest (str): the destination to which the raw file will be saved

    Returns:
        pd.Series: multindexed series of xCO2 with (time, lat) as coords.
    """
    import re

    from pathlib import Path

    import numpy as np
    import pandas as pd
    import pooch

    # save to temporary location with pooch
    print(
        f"[SeaFlux] Downloading {noaa_mbl_url} to {dest} and reading in as pd.DataFrame"
    )

    dest = Path(dest)
    fname = pooch.retrieve(
        url=noaa_mbl_url,
        known_hash=None,
        path=str(dest.parent),
        fname=str(dest.name),
    )

    # find start line
    is_mbl_surface = False
    for start_line, line in enumerate(open(fname)):
        if re.findall("MBL.*SURFACE", line):
            is_mbl_surface = True
        if not line.startswith("#"):
            break
    if not is_mbl_surface:
        raise Exception(
            "The file at the provided url is not an MBL SURFACE file. "
            "Please check that you have provided the surface url. "
        )

    # read fixed width file CO2
    df = pd.read_fwf(fname, skiprows=start_line, header=None, index_col=0)
    df.index.name = "date"
    # every second line is uncertainty
    df = df.iloc[:, ::2]
    # latitude is given as sin(lat)
    df.columns = np.rad2deg(np.arcsin(np.linspace(-1, 1, 41)))

    # resolve time properly
    year = (df.index.values - (df.index.values % 1)).astype(int)
    day_of_year = ((df.index.values - year) * 365 + 1).astype(int)
    date_strings = ["{}-{:03d}".format(*a) for a in zip(year, day_of_year)]
    date = pd.to_datetime(date_strings, format="%Y-%j")
    df = df.set_index(date)
    df = df.iloc[:-1]  # remove the last value that is for 2020-01-01

    # renaming indexes (have to stack for that)
    df = df.stack()
    index = df.index.set_names(["time", "lat"])
    df = df.set_axis(index)

    df.source = noaa_mbl_url

    return df


def download_noaa_mbl(
    noaa_mbl_url,
    download_dest="../data/raw/co2_GHGreference_surface.txt",
    target_lat=None,
    target_lon=None,
    interp_method="linear",
):
    """
    Downloads the NOAA marine boundary layer xCO2 and grids it
    to a defined grid if target lat and lon provided.

    Args:
        noaa_mbl_url (str): surface xCO2 file from https://www.esrl.noaa.gov/gmd/ccgg/mbl/index.html
        target_lat (None, array_like): if None, the default lats will be returned,
        if array-like then will return xCO2 interpolated onto the given latitudes
        target_lon (None, array-like): if None, data will not be broadcast (expanded) along
        latitudes, if array-like, then will broadcast to those longitudes
        inter_method (str): can be linear or nearest

    Returns:
        xr.DataArray: the data array contains xCO2 interpolated onto the given lats and lons
    """
    import numpy as np
    import xarray as xr

    from pandas import Timestamp

    history = (
        f"[SeaFlux@{Timestamp.today():%Y-%m-%dT%H:%M}]: "
        f"downloaded NOAA MBL data from {noaa_mbl_url}, "
    )

    df = read_noaa_mbl_url(noaa_mbl_url, download_dest)

    print("[SeaFlux] Converting pd.DataFrame to xr.DataArray")
    xda = df.to_xarray()

    if target_lat is not None:
        history += f"latitude interpolated with {interp_method}, "
        xda = xda.interp(lat=target_lat, method=interp_method)

    if target_lon is not None:
        history += "longitude broadcast"
        lon = xr.DataArray(np.ones_like(target_lon), dims=["lon"], coords=[target_lon])
        xda = xda * lon

    xda.attrs = dict(
        units="ppm",
        product="NOAA Greenhouse Gas Marine Boundary Layer Reference",
        history=history,
        source="https://www.esrl.noaa.gov/gmd/ccgg/mbl/index.html",
        description=(
            "mole fraction of CO2 for the marine boundary layer varying by "
            "latitude and time. Note that values are constant along "
            "longitudes. "
        ),
    )

    return xda


def interpolate_year(co2_dataarray):
    """Interpolates atmospheric pCO2 based on average increases

    The data is not truly interpolated, but rather follows the
    method shown below.

    1. [anom] calculate climatological anomaly (clim - mean)
    2. [diff] calculate average difference from DEC to JAN
    3. [last] get last value of last full year of CO2

    Args:
        co2_dataarray (xr.DataArray): a data array of monthly pCO2atm
        year (int, float): the year to be interpolated, typically, the last

    Returns:
        xr.DataArray: the input array + 'interpolated' data with added meta
    """
    import pandas as pd
    import xarray as xr

    print("[SeaFlux] Interpolating missing year from previous years")

    co2 = co2_dataarray

    t0, t1 = [
        str(s)
        for s in (
            co2.groupby("time.year")
            .count("time")
            .where(lambda a: a != 0, drop=True)
            .year.values[[0, -1]]
            .astype(str)
        )
    ]
    co2 = co2.sel(time=slice(t0, t1))

    t0 = int(t1) + 1
    t1 = t0 + 1
    time = pd.date_range(f"{t0}", f"{t1}", freq="1MS", closed="left")
    time += pd.Timedelta(14, "D")

    anom = co2.groupby("time.month").mean("time") - co2.mean("time")
    diff = co2.diff("time")[::12].mean("time")
    last = co2[-1].drop("time")

    intp = (
        (last + diff + anom)
        .rename(month="time")
        .assign_coords(time=time)
        .transpose("time", "lat", "lon")
    )

    today = pd.Timestamp.today()
    history = (
        f"[SeaFlux @ {today:%Y-%m-%d}] The year {t0} has been "
        "interpolated by adding the climatological monthly values "
        "to the last value of the NOAA MBL time series. An offset "
        "between JAN and DEC has also been added to the "
        "climatological anomaly."
    )

    history = "; ".join([co2.attrs.get("history", ""), history])
    co2 = xr.concat([co2, intp], "time").assign_attrs(history=history)

    return co2



"""
Gas Transfer Velocity
---------------------

Modulates the magnitude of the flux between the atmosphere and the ocean.
"""

from . import check_units as check


def kw_scaled(wind_speed, wind_grid_stdev, temp_C, ice_frac, scaling=16):
    """
    Calculates kw using the wind^2 formulation for different wind products and
    scales them according to a given global 14C bomb value.
    kw = a * <U^2> * (Sc / 660)^-0.5     (1)

    If kw for the wind product and resolution has already been calculated,
    then this step does not have to be performed. Else, the downscaling and
    wind product used will have an influence (Sweeney et al. 2007). Unlike in
    Sweeney et al. (2007), we assume that downscaled temperature will not
    have a large impact on the Schmidt number - this assumption breaks down
    when the time period is too long.

    Using formulation 1 we can calculate a:
    a = scaled / (<U^2> * (Sc / 660)^-0.5 * (1 - ice_frac))
    kw = a * <U^2> * (Sc / 660)^-0.5

    Parameters
    ----------
    wind_speed : np.array
        wind speed in m/s
    wind_grid_stdev : np.array
        variance of the wind speed in each grid cell for resampled data
    temp_C : np.array
        temperature in degrees C
    ice_frac : np.array
        the sea ice cover as a fraction
    scaling : int [16 cm/hr]
        the global estimate of 14C bomb flux. The default (16 cm/hr) is from
        Wanninkhof (2013). But Sweeney et al. (2007) suggest 14.6 cm/hr

    Returns
    -------
    kw_scaled : np.array
        gas transfer velocity in cm/hr
    attributes : dict
        a dictionary containing information about the calculation to be used
        for creating netCDF files

    """
    from numpy import around, array, isnan, nan_to_num, nansum

    def calculate_scaled_alpha(U2, Sc, ice, scaling):
        """scales alpha"""
        mask = isnan(U2) | isnan(Sc)

        weight = nan_to_num(1 - ice)
        weight[mask] = 0

        a = around(scaling / (nansum(U2 * Sc * weight) / nansum(weight)), 4)

        return a

    Uavg = array(wind_speed)
    Ustd = array(wind_grid_stdev)
    T = array(temp_C)
    ice = array(ice_frac)

    Uavg2 = Uavg ** 2
    U2 = Uavg2 + Ustd ** 2

    Sc = schmidt_number(T)
    Sc660 = (Sc / 660) ** -0.5

    a = calculate_scaled_alpha(U2, Sc660, ice, scaling)

    k = a * Uavg2 * Sc660

    attrs = {
        "long_name": "Gas transfer velocity for CO2",
        "units": "cm/hr",
        "alpha": a,
        "description": (
            f"k = a * <U2> * Sc660 where a is scaled to a "
            f"global 14C bomb estimate of {scaling} cm/hr. "
            f"Note that the global kw is weighted by sea ice "
            f"cover, thus the mean of kw will not be "
            f"{scaling} cm/hr. "
            f"<U2> is the second moment of the wind speed "
            f"calculated as <Uavg^2 + Ustd^2>. "
            f"Sc660 is the Schmidt number for CO2 normalised "
            f"to 660 as in Wanninkhof (1992): (Sc / 660)^-0.5. "
            f"For more details on this calculation see "
            f"Sarmiento and Gruber (2006), Wanninkhof (2014) "
            f"and Sweeney et al. (2007). "
        ),
    }
    return k, attrs


def schmidt_number(temp_C):
    """
    Calculates the Schmidt number as defined by Jahne et al. (1987) and listed
    in Wanninkhof (2014) Table 1.

    Args:
        temp_C (array): temperature in degrees C

    Returns:
        array: Schmidt number (dimensionless)

    Examples:
        >>> schmidt_number(20)  # from Wanninkhof (2014)
        668.344

    References:
        Jähne, B., Heinz, G., & Dietrich, W. (1987). Measurement of the
        diffusion coefficients of sparingly soluble gases in water. Journal
        of Geophysical Research: Oceans, 92(C10), 10767–10776.
        https://doi.org/10.1029/JC092iC10p10767
    """

    check.temp_K(temp_C + 273.15)
    T = temp_C

    a = +2116.8
    b = -136.25
    c = +4.7353
    d = -0.092307
    e = +0.0007555

    Sc = a + b * T + c * T ** 2 + d * T ** 3 + e * T ** 4

    return Sc


def k_Li86(wind_ms, temp_C):
    """
    Calculates the gas transfer coeffcient for CO2 using the formulation
    of Liss and Merlivat (1986)

    Note:
        This is an old parameterization and we recommend using updated
        parameterisations that are calculated based on the wind product you
        choose to use. We include this parameterisation based purely for
        legacy purposes.

    Args:
        wind_ms (array): wind speed in m/s
        temp_C (array): temperature in degrees C

    Returns:
        kw (array): gas transfer velocity (k600) in cm/hr

    References:
        Liss, P. S., & Merlivat, L. (1986). The Role of Air-Sea Exchange
        in Geochemical Cycling (Vol. 1983, Issue June 1983).
        D. Reidel Publishing Company.
    """
    from numpy import array, zeros_like

    U = array(wind_ms)
    T = array(temp_C)

    check.temp_K(T + 273.15)
    check.wind_ms(U)

    Sc = schmidt_number(T)
    k = zeros_like(temp_C)

    i1 = U <= 3.6
    i2 = (U > 3.6) & (U < 13.0)
    i3 = U >= 13.0

    k[i1] = (0.17 * U[i1]) * (Sc[i1] / 600) ** (-2.0 / 3.0)
    k[i2] = ((U[i2] - 3.4) * 2.8) * (600 / Sc[i2]) ** 0.5
    k[i3] = ((U[i3] - 8.4) * 5.9) * (600 / Sc[i3]) ** 0.5

    return k


def k_Wa92(wind_second_moment, temp_C):
    """
    Calculates the gas transfer coeffcient for CO2 using the formulation
    of Wanninkhof (1992)

    Note:
        This is an old parameterization and we recommend using updated
        parameterisations that are calculated based on the wind product you
        choose to use. We include this parameterisation based purely for
        legacy purposes.

    The gas transfer velocity is scaled from instantaneous wind speeds.
    The study applies a correction to the scaling (0.39) based on instantaneous
    wind speeds to lower it to 0.31. This correction is based on the variability
    of wind.

    .. math::
        k_{660} = 0.31 \\cdot U^2


    Args:
        wind_second_moment (array): wind speed squared in m2/s2. Note that the
        second moment should be calculated at the native resolution of the
        wind to avoid losses of variability when taking the square product.
        temp_C (array): temperature in degrees C

    Returns:
        kw (array): gas transfer velocity (k660) in cm/hr

    References:
        Wanninkhof, R. H. (1992). Relationship between wind speed and gas
        exchange over the ocean. Journal of Geophysical Research, 97(C5),
        7373. https://doi.org/10.1029/92JC00188
    """
    from numpy import array

    U2 = array(wind_second_moment)
    T = array(temp_C)

    check.temp_K(T + 273.15)

    Sc = schmidt_number(temp_C)
    k = (0.31 * U2) * (660 / Sc) ** 0.5

    return k


def k_Wa99(wind_ms, temp_C):
    """
    Calculates the gas transfer coeffcient for CO2 using the formulation
    of Wanninkhof and McGillis (1999)

    The gas transfer velocity has been scaled for in-situ short term wind
    products. Note that using this function for any other wind product is not
    correct.

    .. math::
        k_{600} = 0.0283 \\cdot U^3

    Args:
        wind_ms (array): wind speed in m/s
        temp_C (array): temperature in degrees C

    Returns:
        kw (array): gas transfer velocity (k600) in cm/hr

    References:
        Wanninkhof, R. H., & McGillis, W. R. (1999). A cubic relationship
        between air-sea CO2 exchange and wind speed. Geophysical Research
        Letters, 26(13), 1889–1892. https://doi.org/10.1029/1999GL900363
    """
    from numpy import array

    U = array(wind_ms)
    T = array(temp_C)

    check.temp_K(T + 273.15)
    check.wind_ms(U)

    Sc = schmidt_number(temp_C)
    k = (0.0283 * U ** 3) * (600 / Sc) ** 0.5

    return k


def k_Ni00(wind_ms, temp_C):
    """
    Calculates the gas transfer coeffcient for CO2 using the formulation
    of Nightingale et al (2000)

    .. math::
        k_{600} = 0.333 \\cdot U + 0.222 \\cdot U^2

    Args:
        wind_ms (array): wind speed in m/s
        temp_C (array): temperature in degrees C

    Returns:
        kw (array): gas transfer velocity (k600) in cm/hr

    References:
        Nightingale, P. D., Malin, G., Law, C. S., Watson, A. J., Liss, P. S.,
        Liddicoat, M. I., Boutin, J., & Upstill-Goddard, R. C. (2000). In
        situ evaluation of air-sea gas exchange parameterizations using
        novel conservative and volatile tracers. In Global Biogeochemical
        Cycles (Vol. 14, Issue 1, p. 373). https://doi.org/10.1029/1999GB900091
    """
    from numpy import array

    U = array(wind_ms)
    T = array(temp_C)

    check.temp_K(T + 273.15)
    check.wind_ms(U)

    Sc = schmidt_number(temp_C)
    k = (0.333 * U + 0.222 * U ** 2) * (600 / Sc) ** 0.5

    return k


def k_Mc01(wind_ms, temp_C):
    """
    Calculates the gas transfer coeffcient for CO2 using the formulation
    of McGillis et al. (2001)

    The gas transfer velocity has been scaled for in-situ short term wind
    products. Note that using this function for any other wind product is not
    correct.

    .. math::
        k_{660} = 3.3 + 0.026 \\cdot U^3

    Args:
        wind_ms (array): wind speed in m/s
        temp_C (array): temperature in degrees C

    Returns:
        kw (array): gas transfer velocity (k660) in cm/hr

    References:
        McGillis, W. R., Edson, J. B., Ware, J. D., Dacey, J. W. H., Hare, J. E.,
        Fairall, C. W., & Wanninkhof, R. H. (2001). Carbon dioxide flux
        techniques performed during GasEx-98. Marine Chemistry, 75(4), 267–280.
        https://doi.org/10.1016/S0304-4203(01)00042-1
    """
    from numpy import array

    U = array(wind_ms)
    T = array(temp_C)

    check.temp_K(T + 273.15)
    check.wind_ms(U)

    Sc = schmidt_number(temp_C)
    k = 3.3 + (0.026 * U ** 3) * (660 / Sc) ** 0.5

    return k


def k_Ho06(wind_second_moment, temp_C):
    """
    Calculates the gas transfer coeffcient for CO2 using the formulation
    of Ho et al. (2006)

    The gas transfer velocity is for the QuickSCAT satellite wind product.
    Note that using this function for any other wind product is stricktly
    speaking not correct.

    .. math::
        k_{600} = 0.266 \\cdot U^2

    The parameterization is based on the SOLAS Air-Sea Gas Exchange (SAGE)
    experiment.

    Args:
        wind_ms (array): wind speed in m/s
        temp_C (array): temperature in degrees C

    Returns:
        kw (array): gas transfer velocity (k600) in cm/hr

    References:
        Ho, D. T., Law, C. S., Smith, M. J., Schlosser, P., Harvey, M., & Hill, P.
        (2006). Measurements of air-sea gas exchange at high wind speeds in the Southern
        Ocean: Implications for global parameterizations. Geophysical Research Letters,
        33(16), 1–6. https://doi.org/10.1029/2006GL026817
    """
    from numpy import array

    U2 = array(wind_second_moment)
    T = array(temp_C)

    check.temp_K(T + 273.15)

    Sc = schmidt_number(temp_C)
    k = (0.266 * U2) * (600 / Sc) ** 0.5

    return k


def k_Sw07(wind_second_moment, temp_C):
    """
    Calculates the gas transfer coeffcient for CO2 using the formulation
    Wanninkhof (1992) rescaled by Sweeny et al (2007)

    The gas transfer velocity has been scaled for the NCEP/NCAR reanalysis 1
    product. Note that using this function for any other wind product is not
    correct.

    .. math::
        k_{660} = 0.27 \\cdot U^2

    Args:
        wind_second_moment (array): wind speed squared in m2/s2. Note that the
            second moment should be calculated at the native resolution of the
            wind to avoid losses of variability when taking the square product.
        temp_C (array): temperature in degrees C

    Returns:
        kw (array): gas transfer velocity (k660) in cm/hr

    References:
        Sweeney, C., Gloor, E., Jacobson, A. R., Key, R. M., McKinley, G. A.,
        Sarmiento, J. L., & Wanninkhof, R. H. (2007). Constraining global
        air-sea gas exchange for CO2 with recent bomb 14C measurements.
        Global Biogeochemical Cycles, 21(2). https://doi.org/10.1029/2006GB002784
    """
    from numpy import array

    U2 = array(wind_second_moment)
    T = array(temp_C)

    check.temp_K(T + 273.15)

    Sc = schmidt_number(temp_C)
    k = (0.27 * U2) * (660 / Sc) ** 0.5

    return k


def k_Wa09(wind_ms, temp_C):
    """
    Calculates the gas transfer coeffcient for CO2 using the formulation
    of Wanninkhof et al. (2009)

    The gas transfer velocity has been scaled for the Cross-Calibrated Multi-
    Platform (CCMP) Winds product. Note that using this function for any other
    wind product is not correct.

    .. math::
        k_{660} = 3.0 + 0.1 \\cdot U + 0.064 \\cdot U^2 + 0.011 \\cdot U^3

    Args:
        wind_ms (array): wind speed in m/s
        temp_C (array): temperature in degrees C

    Returns:
        kw (array): gas transfer velocity (k660) in cm/hr

    References:
        Wanninkhof, R. H., Asher, W. E., Ho, D. T., Sweeney, C., & McGillis,
        W. R. (2009). Advances in Quantifying Air-Sea Gas Exchange and
        Environmental Forcing*. Annual Review of Marine Science, 1(1),
        213–244. https://doi.org/10.1146/annurev.marine.010908.163742
    """
    from numpy import array

    U = array(wind_ms)
    T = array(temp_C)

    check.temp_K(T + 273.15)
    check.wind_ms(U)

    Sc = schmidt_number(temp_C)
    k = (3.0 + 0.1 * U + 0.064 * U ** 2 + 0.011 * U ** 3) * (660 / Sc) ** 0.5

    return k


def k_Wa14(wind_second_moment, temp_C):
    """
    Calculates the gas transfer coeffcient for CO2 using the formulation
    of Wanninkhof et al. (2014)

    The gas transfer velocity has been scaled for the Cross-Calibrated Multi-
    Platform (CCMP) Winds product. Note that using this function for any other
    wind product is not correct.

    .. math::
        k_{660} = 0.251 \\cdot U^2

    Args:
        wind_second_moment (array): wind speed squared in m2/s2. Note that the
            second moment should be calculated at the native resolution of the
            wind to avoid losses of variability when taking the square product.
        temp_C (array): temperature in degrees C

    Returns:
        kw (array): gas transfer velocity (k660) in cm/hr

    References:
        Wanninkhof, R. H. (2014). Relationship between wind speed and gas
        exchange over the ocean revisited. Limnology and Oceanography:
        Methods, 12(JUN), 351–362. https://doi.org/10.4319/lom.2014.12.351
    """
    from numpy import array

    U2 = array(wind_second_moment)
    T = array(temp_C)

    check.temp_K(T + 273.15)

    Sc = schmidt_number(temp_C)
    k = 0.251 * U2 * (660 / Sc) ** 0.5

    return k
